package ac.essex.ooechs.imaging.commons.signal;

import ac.essex.ooechs.imaging.commons.PixelLoader;
import ac.essex.ooechs.imaging.commons.ColourChannels;
import ac.essex.ooechs.imaging.commons.StatisticsSolver;
import ac.essex.ooechs.imaging.commons.util.panels.ImageFrame;
import ac.essex.ooechs.imaging.commons.fast.FastStatistics;
import ac.essex.ooechs.imaging.commons.window.data.Window;

import java.util.Vector;
import java.awt.image.BufferedImage;

import org.jfree.data.category.DefaultCategoryDataset;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.plot.PlotOrientation;


/**
 * Converts an image into various signals. This means converting
 * the image into a a series of points, for different values of X.
 * <p/>
 * The points can be calculated based on average intensity of a particular
 * channel, or other statistics.
 *
 * @author Olly Oechsle, University of Essex, Date: 16-Apr-2008
 * @version 1.0
 */
public class Signal {

    public double[] data;

    public int channel, statistic;

    public int classID;

    public static void main(String[] args) throws Exception {
        //PixelLoader image = new PixelLoader("/home/ooechs/Desktop/pipes_coloured/Foreground/image121.bmp");
        PixelLoader image = new PixelLoader("/home/ooechs/Desktop/pipes_coloured/Unusual/image128.bmp");
        Signal s1 = new Signal(ColourChannels.GREYSCALE, StatisticsSolver.MEAN, image);
        new ImageFrame(s1.getChart(400, 300), "Mean: Grey");

        Signal s1r = new Signal(ColourChannels.RED, StatisticsSolver.MEAN, image);
        new ImageFrame(s1r.getChart(400, 300), "Red");

        Signal s1g = new Signal(ColourChannels.GREEN, StatisticsSolver.MEAN, image);
        new ImageFrame(s1g.getChart(400, 300), "Green");


        Signal s1b = new Signal(ColourChannels.BLUE, StatisticsSolver.MEAN, image);
        new ImageFrame(s1b.getChart(400, 300), "Blue");

        Signal s2 = new Signal(0, StatisticsSolver.STANDARD_DEVIATION, image);
        new ImageFrame(s2.getChart(400, 300), "Std Dev: Grey");
    }

    public Signal(SignalDefinition d, PixelLoader image) {
        this(d.channel, d.statistic, image);
    }
    
    public Signal(int channel, int statistic, PixelLoader image) {
        this(channel, statistic, image, new Window(image.getWidth(), image.getHeight(), 0, 0, null));
    }

    public Signal(SignalDefinition d, PixelLoader image, Window w) {
        this(d.channel, d.statistic, image, w);
    }

    public Signal(int channel, int statistic, PixelLoader image, Window window) {

        // one value for each value of x
        data = new double[window.getWidth()];

        this.channel = channel;
        this.statistic = statistic;

        data = getSignal(image, window);

    }
    

    private double[] getSignal(PixelLoader image, Window window) {

        Vector<Double> values = new Vector<Double>();

        int blockSize = 1;

        // move along the window one pixel at a time and get the summed intensity
        for (int wx = 0; wx < window.width - blockSize; wx += blockSize) {

            FastStatistics total = new FastStatistics();
            int x = wx + window.x;
            for (int wy = 0; wy < window.height; wy++) {
                int y = wy + window.y;
                switch (channel) {
                    case ColourChannels.GREYSCALE:
                        total.addData(image.getGreyValue(x, y));
                        break;
                    case ColourChannels.RED:
                        total.addData(image.getRed(x, y));
                        break;
                    case ColourChannels.GREEN:
                        total.addData(image.getGreen(x, y));
                        break;
                    case ColourChannels.BLUE:
                        total.addData(image.getBlue(x, y));
                        break;
                }

            }
            values.add((double) total.getStatistic(statistic));
        }

        // convert vector into an array
        double[] v = new double[values.size()];
        for (int i = 0; i < values.size(); i++) {
            Double aDouble = values.elementAt(i);
            v[i] = aDouble;
        }

        return v;

    }

    public BufferedImage getChart(int width, int height) {
        DefaultCategoryDataset series = new DefaultCategoryDataset();
        for (int x = 0; x < data.length; x++) {
            double value = data[x];
            series.addValue(value, "i", String.valueOf(x));
        }
        JFreeChart myChart = ChartFactory.createLineChart(null, "x", "y", series, PlotOrientation.VERTICAL, false, false, false);
        return myChart.createBufferedImage(width, height);
    }

}
