package ac.essex.ooechs.imaging.commons.watershed;

import ac.essex.ooechs.imaging.commons.PixelLoader;
import ac.essex.ooechs.imaging.commons.Pixel;

import java.util.Vector;
import java.util.Stack;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.File;

/**
 * Performs the vincent soille algorithm on an image.
 */
public class WatershedAlgorithm {

    public static void main(String[] args) throws Exception {
        //
        //WatershedAlgorithm s = new WatershedAlgorithm(new PixelLoader("/home/ooechs/Desktop/balloon.jpg"));
        WatershedAlgorithm s = new WatershedAlgorithm(new PixelLoader("/home/ooechs/Desktop/unionjack1.jpg"));
        s.process(10);
        //s.draw("/home/ooechs/Desktop/watershed2.jpg");
    }

    // The image to work on
    protected PixelLoader image;

    // The groups that will be created
    public Vector<PixelGroup> groups;

    public int minGroupSize;

    protected int data[][];

    protected int groupIDCounter = 1;

    public WatershedAlgorithm(PixelLoader image) {

        this.image = image;

        // create an array to remember stuff with
        data = new int[image.getWidth()][image.getHeight()];

        // make a border around the data so that array index out of bounds exceptions can be avoided
        int h = image.getHeight()-1;
        for (int x = 0; x < image.getWidth(); x++)  {
            data[x][0] = -1;
            data[x][h] = -1;
        }
        
        int w = image.getWidth() - 1;
        for (int y = 0; y < image.getHeight(); y++)  {
            data[0][y] = -1;
            data[w][y] = -1;
        }

        groups = new Vector<PixelGroup>(10);

        minGroupSize = (int) ((image.getWidth() * image.getHeight()) * 0.001);

        System.out.println("Min Group Size: " + minGroupSize);

    }

    public void process(int floodStep) {

        System.out.println("Churning Watershed Algorithm");

        long start = System.currentTimeMillis();

        for (int threshold = floodStep; threshold < 255; threshold += floodStep) {

            if (groups.size() > 0) {
                for (int j = 0; j < groups.size(); j++) {
                    PixelGroup group = groups.elementAt(j);
                    grow(group, threshold);
                }
            }

            findNewGroups(threshold);

            draw("/home/ooechs/Desktop/frame" + threshold + ".png");

        }

        long time = System.currentTimeMillis() - start;

        System.out.println("Processed in: " + time + "ms.");

        System.out.println("Groups: " + groups.size());

    }

    public void findNewGroups(int threshold) {

        // find all pixels at given threshold
        for (int y = 1; y < image.getHeight() - 2; y++) {
            for (int x = 1; x < image.getWidth() - 2; x++) {

                if (data[x][y] == 0) {

                    // get this pixel's colour
                    int pixelColour = image.getGreyValue(x, y);

                    // is this pixel under the threshold
                    if (pixelColour <= threshold) {
                        PixelGroup g = new PixelGroup(groupIDCounter, x, y);
                        data[x][y] = groupIDCounter;
                        expand(x, y, threshold, g);
                        // ensure the group is not too small
                        if (g.pixels.size() > minGroupSize) {
                            groups.add(g);
                            groupIDCounter++;
                        } else {
                            // the group is too small - undo the changes
                            for (int i = 0; i < g.pixels.size(); i++) {
                                Pixel pixel =  g.pixels.elementAt(i);
                                data[pixel.x][pixel.y] = 0;
                            }
                        }
                    }

                }

            }
        }

    }

    public void grow(PixelGroup g, int threshold) {
        for (int i = 0; i < g.pixels.size(); i++) {
            Pixel pixel = g.pixels.elementAt(i);
            expand(pixel.x, pixel.y, threshold, g);
        }
    }

    public void expand(int x, int y, int threshold, PixelGroup g) {
        Stack<Pixel> toExpand = new Stack<Pixel>();
        toExpand.push(new Pixel(x, y));
        while (toExpand.size() > 0) {
            Pixel p = toExpand.pop();
            expand(p.x, p.y, threshold, g, toExpand);
        }
    }

    private void expand(int x, int y, int threshold, PixelGroup g, Stack<Pixel> toExpand) {
        for (int dy = -1; dy <= 1; dy++)
        for (int dx = -1; dx <= 1; dx++) {
            // don't allow diagonal connections
            if (dx != 0 && dy != 0) continue;
            if (data[x+dx][y+dy] == 0 && image.getGreyValue(x+dx,y+dy) <= threshold) {
                toExpand.push(g.add(x+dx, y+dy));
            }
        }
    }

    private boolean isEdgePixel(int x, int y, int pixelClass) {
        return data[x][y - 1] != pixelClass || data[x - 1][y] != pixelClass || data[x][y + 1] != pixelClass || data[x + 1][y] != pixelClass;
    }

    public void draw(String filename) {

        int red = new Color(255, 0, 0).getRGB();
        int green = new Color(0, 255, 0).getRGB();
        int blue = new Color(0, 0, 255).getRGB();
        int yellow = new Color(255, 255, 0).getRGB();
        int darkBlue = new Color(45, 45, 100).getRGB();
        int orange = new Color(255, 128, 30).getRGB();

        int white = Color.BLACK.getRGB();

        int[] colours = new int[]{red, green, blue, yellow, orange, darkBlue};

        try {

            //PixelLoader img = new PixelLoader(image.getFile());
            BufferedImage img = new BufferedImage(image.getWidth(), image.getHeight(), BufferedImage.TYPE_INT_RGB);
            Graphics g = img.getGraphics();
            g.setColor(Color.WHITE);
            g.fillRect(0,0,image.getWidth(), image.getHeight());

            int currentColour = 0;

            int pixelCount = 0;

            for (int i = 0; i < groups.size(); i++) {
                PixelGroup group = groups.elementAt(i);

                // pick a colour any colour
                if (currentColour >= colours.length) currentColour = 0;

                // get the group pixels
                Vector<Pixel> pixels = group.pixels;

                for (int j = 0; j < pixels.size(); j++) {
                    Pixel pixel = pixels.elementAt(j);
                    if (isEdgePixel(pixel.x, pixel.y, group.groupID)) {
                        img.setRGB(pixel.x, pixel.y, white);
                    } else {
                        img.setRGB(pixel.x, pixel.y, colours[currentColour]);
                        //img.setRGB(pixel.x, pixel.y, blue);
                    }
                    pixelCount++;
                }

                currentColour++;

            }

            System.out.println("Drawn: " + pixelCount + " pixels.");

            //img.saveAs(filename);
            javax.imageio.ImageIO.write(img, "png", new File(filename));

        } catch (Exception e) {
            e.printStackTrace();
        }
    }


    public class PixelGroup {

        protected Vector<Pixel> pixels;
        protected int groupID;

        public PixelGroup(int groupID, int x, int y) {
            this.groupID = groupID;
            this.pixels = new Vector<Pixel>(10);
            add(x, y);
        }

        public Pixel add(int x, int y) {
            data[x][y] = groupID;
            Pixel p = new Pixel(x, y);
            pixels.add(p);
            return p;
        }

    }

}
